from typing import Callable, Dict, List, Any

import re

import numpy as np
import pandas as pd
from alibi.explainers import KernelShap
from bs4 import BeautifulSoup

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel


def explain(model: GiskardModel, dataset: GiskardDataset, input_data: Dict):
    df = model.prepare_dataframe(dataset)
    feature_names = list(df.columns)

    # Make sure column order is that column order is the same as in df
    input_data_df = pd.DataFrame([input_data])[df.columns]
    input_df = model.prepare_dataframe(
        GiskardDataset(
            df=input_data_df,
            target=dataset.target,
            feature_types=dataset.feature_types,
            column_types=dataset.column_types,
        )
    )

    def predict_array(array):
        return model.prediction_function(pd.DataFrame(array, columns=list(df.columns)))

    kernel_shap = KernelShap(
        predictor=predict_array,
        feature_names=feature_names,
        task=model.model_type,
    )

    example = background_example(df, dataset.feature_types)
    kernel_shap.fit(example)
    explanations = kernel_shap.explain(input_df)

    if model.model_type == "regression":
        explanation_chart_data = summary_shap_regression(
            shap_values=explanations.shap_values, feature_names=feature_names
        )
    elif model.model_type == "classification":
        explanation_chart_data = summary_shap_classification(
            shap_values=explanations.shap_values,
            feature_names=feature_names,
            class_names=model.classification_labels,
        )
    else:
        raise ValueError(f"Prediction task is not supported: {model.model_type}")
    return explanation_chart_data


def background_example(df: pd.DataFrame, input_types: Dict[str, str]) -> pd.DataFrame:
    example = df.mode(dropna=False).iloc[[0]]  # si plusieurs modes, on prend le premier
    example.fillna("", inplace=True)
    median = df.median()
    num_columns = [key for key in list(df.columns) if input_types.get(key) == "numeric"]
    for column in num_columns:
        example[column] = median[column]
    return example.astype(df.dtypes)


def summary_shap_classification(
        shap_values: List[np.ndarray],
        feature_names: List[str],
        class_names: List[str],
        max_display: int = 5,
) -> Dict[str, Dict[str, Dict[str, float]]]:
    feature_order = np.argsort(np.sum(np.mean(np.abs(shap_values), axis=1), axis=0))
    feature_order = feature_order[-min(max_display, len(feature_order)):]
    feature_inds = feature_order[:max_display]
    chart_data: Dict[str, Dict[Any, Any]] = {"explanations": {}}
    for i in range(len(shap_values)):
        global_shap_values = np.abs(shap_values[i]).mean(0)
        chart_data["explanations"][class_names[i]] = {
            feature_names[feature_ind]: global_shap_values[feature_ind]
            for feature_ind in feature_inds
        }
    return chart_data


def summary_shap_regression(
        shap_values: List[np.ndarray], feature_names: List[str], max_display: int = 5
) -> Dict[str, Dict[str, Dict[str, float]]]:
    feature_order = np.argsort(np.sum(np.mean(np.abs(shap_values), axis=1), axis=0))
    feature_order = feature_order[-min(max_display, len(feature_order)):]
    feature_inds = feature_order[:max_display]
    global_shap_values = np.abs(shap_values[0]).mean(0)
    chart_data = {"explanations": {
        "default": {
            feature_names[feature_ind]: global_shap_values[feature_ind]
            for feature_ind in feature_inds
        }
    }}
    return chart_data


def text_explanation_prediction_wrapper(
        prediction_function: Callable, input_example: pd.DataFrame, text_column: str
) -> Callable:
    def text_predict(text_documents: List[str]):
        num_documents = len(text_documents)
        df_with_text_documents = input_example.append(
            [input_example] * (num_documents - 1), ignore_index=True
        )
        df_with_text_documents[text_column] = pd.DataFrame(text_documents)
        return prediction_function(df_with_text_documents)

    return text_predict


def parse_text_explainer_response(response: str) -> Dict[str, str]:
    text_explanation_soup = BeautifulSoup(response, "html.parser")
    labels = []
    explanations_html = []
    for i, paragraph in enumerate(text_explanation_soup.find_all("p")):
        if (i % 2) == 0:
            label = re.findall(r"\by=.*\b", str(paragraph.find("b")))[0][2:]
            labels.append(label)
        else:
            explanations_html.append(str(paragraph))
    return dict(zip(labels, explanations_html))



import asyncio
import logging
from asyncio import StreamReader, StreamWriter

from tenacity import retry, wait_exponential

from giskard.ml_worker.bridge.error import ConnectionLost
from giskard.ml_worker.bridge.service_messages import CREATE_CLIENT_CHANNEL, START_INNER_SERVER
from giskard.ml_worker.utils.network import readable_hex

logger = logging.getLogger(__name__)


class MLWorkerBridge:
    def __init__(
        self,
        local_port: int,
        remote_host: str,
        remote_port: int,
        execution_loop=asyncio.get_event_loop(),
    ) -> None:
        self.loop = execution_loop

        self.local_host = "localhost"
        self.local_port = local_port

        self.remote_host = remote_host
        self.remote_port = remote_port

    @retry(wait=wait_exponential(min=0.1, max=5, multiplier=0.1))
    async def start(self):
        remote_writer = None
        try:
            logger.info(f"Connecting to {self.remote_host}:{self.remote_port}")
            remote_reader, remote_writer = await asyncio.open_connection(
                self.remote_host, self.remote_port
            )
            logger.info(f"Connected to Giskard server {self.remote_host}:{self.remote_port}")
            await self.send_service_message(remote_writer, START_INNER_SERVER)
            self.loop.create_task(self.listen_remote_server_service_socket(remote_reader))
        except Exception as e:
            remote_writer.close()
            await remote_writer.wait_closed()
            logger.error(
                f"Failed to connect to a remote host: {self.remote_host}:{self.remote_port} : {str(e)}"
            )
            raise e

    async def send_service_message(
        self, remote_writer: StreamWriter, msg_type: int, data: bytes = b""
    ):
        msg = self.create_service_message(msg_type, data)
        logger.debug(f"Service message: Writing {len(msg)} bytes: {readable_hex(msg)}")
        remote_writer.write(msg)
        await remote_writer.drain()

    @staticmethod
    def create_service_message(msg_type: int, data: bytes = b""):
        msg_type_bytes = int.to_bytes(msg_type, 1, "big")
        message_len = len(msg_type_bytes) + len(data)
        msg = message_len.to_bytes(4, "big") + msg_type_bytes + data
        return msg

    async def listen_remote_server_service_socket(self, remote_reader: StreamReader):
        try:
            logger.debug("Created remote server listener task")
            while True:
                logger.debug("waiting for a service command")
                data = await remote_reader.read(9)  # command payload is 1 byte by design
                if len(data):
                    client = data[:8]
                    command = int.from_bytes(data[8:], "big")
                    logger.debug(f"service command received: {client}: {command}")
                    await self.handle_server_command(client, command)
                else:
                    raise ConnectionLost()
        except ConnectionLost:
            logger.info("Connection to Giskard host lost")
            await self.start()
        except Exception as e:
            logger.exception(e)

    async def handle_server_command(self, client, command):
        if command == CREATE_CLIENT_CHANNEL:
            remote_reader, remote_writer = await asyncio.open_connection(
                self.remote_host, self.remote_port
            )
            logger.debug(
                f"Connected client {client} to remote host {(self.remote_host, self.remote_port)}"
            )
            message = self.create_service_message(CREATE_CLIENT_CHANNEL, client)
            logger.debug(
                f"handle_server_command: Writing {len(message)} bytes: {readable_hex(message)}"
            )
            remote_writer.write(message)
            await remote_writer.drain()

            grpc_reader, grpc_writer = await asyncio.open_connection(
                self.local_host, self.local_port
            )
            logger.debug(
                f"Connected client {client} to grpc host {(self.local_host, self.local_port)}"
            )

            self.loop.create_task(
                self.sync_data(
                    client, grpc_reader, remote_writer, f"{client.decode()}: grpc->remote"
                )
            )
            self.loop.create_task(
                self.sync_data(
                    client, remote_reader, grpc_writer, f"{client.decode()}: remote->grpc"
                )
            )

    @staticmethod
    async def sync_data(client, reader: StreamReader, writer: StreamWriter, task_name: str = None):
        log_prefix = "" if not task_name else task_name + ": "
        try:
            try:
                while not reader.at_eof():
                    data = await reader.read(2048)
                    if len(data):
                        logger.debug(f"{log_prefix}Writing {len(data)} bytes: {readable_hex(data)}")
                        writer.write(data)
                        await writer.drain()
                    else:
                        raise ConnectionLost()
            finally:
                writer.close()
        except (ConnectionLost, ConnectionResetError):
            logger.debug(f"{log_prefix}Connection lost: {client}")
        except Exception:
            logger.exception(f"{log_prefix}Sync data error")



import codecs
import socket
from contextlib import closing


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def readable_hex(data):
    s = codecs.encode(data, "hex").decode()
    return " ".join(s[i : i + 2] for i in range(0, len(s), 2))



from typing import Callable, Dict, Optional

import pandas as pd


class GiskardDataset:
    target: str
    feature_types: Dict[str, str]
    column_types: Dict[str, str]
    df: pd.DataFrame

    def __init__(
        self,
        df: pd.DataFrame,
        target: Optional[str],
        feature_types: Dict[str, str],
        column_types: Dict[str, str] = None,
    ) -> None:
        self.df = df
        self.target = target
        self.feature_types = feature_types
        self.column_types = column_types

    @property
    def columns(self):
        return self.df.columns

    def slice(self, slice_fn: Callable):
        if slice_fn is None:
            return self
        return GiskardDataset(slice_fn(self.df), self.target, self.feature_types)

    def __len__(self):
        return len(self.df)



import pandas as pd

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult, TestMessage, TestMessageType
from giskard.ml_worker.testing.abstract_test_collection import AbstractTestCollection
from giskard.ml_worker.testing.utils import apply_perturbation_inplace
from giskard.ml_worker.utils.logging import timer


class MetamorphicTests(AbstractTestCollection):
    @staticmethod
    def _predict_numeric_result(
        ds: GiskardDataset, model: GiskardModel, output_proba=True, classification_label=None
    ):
        if model.model_type == "regression" or not output_proba:
            return model.run_predict(ds).raw_prediction
        elif model.model_type == "classification" and classification_label is not None:
            return model.run_predict(ds).all_predictions[classification_label].values
        elif model.model_type == "classification":
            return model.run_predict(ds).probabilities

    @staticmethod
    def _prediction_ratio(prediction, perturbed_prediction):
        return (
            abs(perturbed_prediction - prediction) / prediction
            if prediction != 0
            else abs(perturbed_prediction)
        )

    @staticmethod
    def _perturb_and_predict(
        ds: GiskardDataset,
        model: GiskardModel,
        perturbation_dict,
        output_proba=True,
        classification_label=None,
    ):
        results_df = pd.DataFrame()
        results_df["prediction"] = MetamorphicTests._predict_numeric_result(
            ds, model, output_proba, classification_label
        )
        modified_rows = apply_perturbation_inplace(ds.df, perturbation_dict)
        if len(modified_rows):
            ds.df = ds.df.iloc[modified_rows]
            results_df = results_df.iloc[modified_rows]
            results_df["perturbed_prediction"] = MetamorphicTests._predict_numeric_result(
                ds, model, output_proba, classification_label
            )
        else:
            results_df["perturbed_prediction"] = results_df["prediction"]
        return results_df, len(modified_rows)

    @timer("Compare and predict the data")
    def _compare_prediction(self, results_df, prediction_task, output_sensitivity=None, flag=None):
        if flag == "Invariance":
            if prediction_task == "classification":
                passed_idx = results_df.loc[
                    results_df["prediction"] == results_df["perturbed_prediction"]
                ].index.values

            elif prediction_task == "regression":
                results_df["predict_difference_ratio"] = results_df.apply(
                    lambda x: self._prediction_ratio(x["prediction"], x["perturbed_prediction"]),
                    axis=1,
                )
                passed_idx = results_df.loc[
                    results_df["predict_difference_ratio"] < output_sensitivity
                ].index.values

        elif flag == "Increasing":
            passed_idx = results_df.loc[
                results_df["prediction"] < results_df["perturbed_prediction"]
            ].index.values

        elif flag == "Decreasing":
            passed_idx = results_df.loc[
                results_df["prediction"] > results_df["perturbed_prediction"]
            ].index.values

        failed_idx = results_df.loc[~results_df.index.isin(passed_idx)].index.values
        return passed_idx, failed_idx

    def _test_metamorphic(
        self,
        flag,
        actual_slice: GiskardDataset,
        model,
        perturbation_dict,
        threshold: float,
        classification_label=None,
        output_sensitivity=None,
        output_proba=True,
    ) -> SingleTestResult:
        actual_slice.df.reset_index(drop=True, inplace=True)

        results_df, modified_rows_count = self._perturb_and_predict(
            actual_slice,
            model,
            perturbation_dict,
            classification_label=classification_label,
            output_proba=output_proba,
        )

        passed_idx, failed_idx = self._compare_prediction(
            results_df, model.model_type, output_sensitivity, flag
        )
        passed_ratio = len(passed_idx) / modified_rows_count if modified_rows_count != 0 else 1

        messages = [
            TestMessage(
                type=TestMessageType.INFO, text=f"{modified_rows_count} rows were perturbed"
            )
        ]

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio > threshold,
                messages=messages,
            )
        )

    def test_metamorphic_invariance(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, output_sensitivity=None
    ) -> SingleTestResult:
        """
        Summary: Tests if the model prediction is invariant when the feature values are perturbed

        Description: -
        For classification: Test if the predicted classification label remains the same after
        feature values perturbation.
        For regression: Check whether the predicted output remains the same at the output_sensibility
        level after feature values perturbation.

        The test is passed when the ratio of invariant rows is higher than the threshold

        Example : The test is passed when, after switching gender from male to female,
        more than 50%(threshold 0.5) of males have unchanged outputs

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            threshold(float):
              Threshold of the ratio of invariant rows
            output_sensitivity(float):
                Optional. The threshold for ratio between the difference between perturbed prediction and actual prediction over
                the actual prediction for a regression model. We consider there is a prediction difference for
                regression if the ratio is above the output_sensitivity of 0.1

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of unchanged rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """

        return self._test_metamorphic(
            flag="Invariance",
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            threshold=threshold,
            output_sensitivity=output_sensitivity,
            output_proba=False,
        )

    def test_metamorphic_increasing(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, classification_label=None
    ):
        """
        Summary: Tests if the model probability increases when the feature values are perturbed

        Description: -
        - For classification: Test if the model probability of a given classification_label is
        increasing after feature values perturbation.

        - For regression: Test if the model prediction is increasing after feature values perturbation.

        The test is passed when the percentage of rows that are increasing is higher than the threshold

        Example : For a credit scoring model, the test is passed when a decrease of wage by 10%,
         default probability is increasing for more than 50% of people in the dataset

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            threshold(float):
              Threshold of the ratio of increasing rows
            classification_label(str):
              Optional.One specific label value from the target column

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of increasing rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """
        assert (
            model.model_type != "classification"
            or str(classification_label) in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic(
            flag="Increasing",
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            threshold=threshold,
        )

    def test_metamorphic_decreasing(
        self, df: GiskardDataset, model, perturbation_dict, threshold=0.5, classification_label=None
    ):
        """
        Summary: Tests if the model probability decreases when the feature values are perturbed

        Description: -
        - For classification: Test if the model probability of a given classification_label is
        decreasing after feature values perturbation.

        - For regression: Test if the model prediction is decreasing after feature values perturbation.

        The test is passed when the percentage of rows that are decreasing is higher than the threshold

        Example : For a credit scoring model, the test is passed when an increase of wage by 10%,
         default probability is decreasing for more than 50% of people in the dataset

        Args:
            df(GiskardDataset):
              Dataset used to compute the test
            model(GiskardModel):
              Model used to compute the test
            perturbation_dict(dict):
              Dictionary of the perturbations. It provides the perturbed features as key
              and a perturbation lambda function as value
            threshold(float):
              Threshold of the ratio of decreasing rows
            classification_label(str):
              Optional. One specific label value from the target column

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            message:
              Test result message
            metric:
              The ratio of decreasing rows over the perturbed rows
            passed:
              TRUE if metric > threshold
        """

        assert (
            model.model_type != "classification"
            or classification_label in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        return self._test_metamorphic(
            flag="Decreasing",
            actual_slice=df,
            model=model,
            perturbation_dict=perturbation_dict,
            classification_label=classification_label,
            threshold=threshold,
        )



import numpy as np
import pandas as pd
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    mean_absolute_error,
    mean_squared_error,
    precision_score,
    r2_score,
    recall_score,
    roc_auc_score,
)

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult
from giskard.ml_worker.testing.abstract_test_collection import AbstractTestCollection


class PerformanceTests(AbstractTestCollection):
    @staticmethod
    def _verify_target_availability(dataset):
        if not dataset.target:
            raise ValueError("Target column is not available")

    def test_auc(self, actual_slice: GiskardDataset, model: GiskardModel, threshold=1.0):

        """
        Test if the model AUC performance is higher than a threshold for a given slice

        Example : The test is passed when the AUC for females is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value of AUC metrics

        Returns:
          actual_slices_size:
              Length of actual_slice tested
          metric:
              The AUC performance metric
          passed:
              TRUE if AUC metrics >= threshold
        """
        self._verify_target_availability(actual_slice)
        if len(model.classification_labels) == 2:
            metric = roc_auc_score(
                actual_slice.df[actual_slice.target], model.run_predict(actual_slice).raw_prediction
            )
        else:
            predictions = model.run_predict(actual_slice).all_predictions
            non_declared_categories = set(predictions.columns) - set(
                actual_slice.df[actual_slice.target].unique()
            )
            assert not len(
                non_declared_categories
            ), f'Predicted classes don\'t exist in the dataset "{actual_slice.target}" column: {non_declared_categories}'

            metric = roc_auc_score(
                actual_slice.df[actual_slice.target], predictions, multi_class="ovr"
            )

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)], metric=metric, passed=metric >= threshold
            )
        )

    def _test_classification_score(
        self, score_fn, gsk_dataset: GiskardDataset, model: GiskardModel, threshold=1.0
    ):
        self._verify_target_availability(gsk_dataset)
        is_binary_classification = len(model.classification_labels) == 2
        gsk_dataset.df.reset_index(drop=True, inplace=True)
        actual_target = gsk_dataset.df[gsk_dataset.target].astype(str)
        prediction = model.run_predict(gsk_dataset).prediction
        if is_binary_classification:
            metric = score_fn(actual_target, prediction, pos_label=model.classification_labels[1])
        else:
            metric = score_fn(actual_target, prediction, average="macro")

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(gsk_dataset)], metric=metric, passed=metric >= threshold
            )
        )

    def _test_accuracy_score(self, gsk_dataset: GiskardDataset, model: GiskardModel, threshold=1.0):
        self._verify_target_availability(gsk_dataset)
        gsk_dataset.df.reset_index(drop=True, inplace=True)
        prediction = model.run_predict(gsk_dataset).prediction
        actual_target = gsk_dataset.df[gsk_dataset.target].astype(str)

        metric = accuracy_score(actual_target, prediction)

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(gsk_dataset)], metric=metric, passed=metric >= threshold
            )
        )

    def _test_regression_score(
        self, score_fn, giskard_ds, model: GiskardModel, threshold=1.0, r2=False
    ):
        results_df = pd.DataFrame()
        giskard_ds.df.reset_index(drop=True, inplace=True)
        self._verify_target_availability(giskard_ds)

        results_df["actual_target"] = giskard_ds.df[giskard_ds.target]
        results_df["prediction"] = model.run_predict(giskard_ds).raw_prediction

        metric = score_fn(results_df["actual_target"], results_df["prediction"])

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(giskard_ds)],
                metric=metric,
                passed=metric >= threshold if r2 else metric <= threshold,
            )
        )

    def test_f1(self, actual_slice: GiskardDataset, model: GiskardModel, threshold=1.0):
        """
        Test if the model F1 score is higher than a defined threshold for a given slice

        Example: The test is passed when F1 score for females is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for F1 Score

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            metric:
              The F1 score metric
            passed:
              TRUE if F1 Score metrics >= threshold
        """
        return self._test_classification_score(f1_score, actual_slice, model, threshold)

    def test_accuracy(self, actual_slice: GiskardDataset, model: GiskardModel, threshold=1.0):
        """
        Test if the model Accuracy is higher than a threshold for a given slice

        Example: The test is passed when the Accuracy for females is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Accuracy

        Returns:
          actual_slices_size:
              Length of actual_slice tested
          metric:
              The Accuracy metric
          passed:
              TRUE if Accuracy metrics >= threshold
        """
        return self._test_accuracy_score(actual_slice, model, threshold)

    def test_precision(self, actual_slice, model: GiskardModel, threshold=1.0):
        """
        Test if the model Precision is higher than a threshold for a given slice

        Example: The test is passed when the Precision for females is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Precision
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            metric:
              The Precision metric
            passed:
              TRUE if Precision metrics >= threshold
        """
        return self._test_classification_score(precision_score, actual_slice, model, threshold)

    def test_recall(self, actual_slice, model: GiskardModel, threshold=1.0):
        """
        Test if the model Recall is higher than a threshold for a given slice

        Example: The test is passed when the Recall for females is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Recall
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            metric:
              The Recall metric
            passed:
              TRUE if Recall metric >= threshold
        """
        return self._test_classification_score(recall_score, actual_slice, model, threshold)

    @staticmethod
    def _get_rmse(y_actual, y_predicted):
        return np.sqrt(mean_squared_error(y_actual, y_predicted))

    def test_rmse(self, actual_slice, model: GiskardModel, threshold=1.0):
        """
        Test if the model RMSE is lower than a threshold

        Example: The test is passed when the RMSE is lower than 10

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for RMSE
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            metric:
              The RMSE metric
            passed:
              TRUE if RMSE metric <= threshold
        """
        return self._test_regression_score(self._get_rmse, actual_slice, model, threshold)

    def test_mae(self, actual_slice, model: GiskardModel, threshold=1.0):
        """
        Test if the model Mean Absolute Error is lower than a threshold

        Example: The test is passed when the MAE is lower than 10

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for MAE

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The MAE metric
            passed:
              TRUE if MAE metric <= threshold
        """
        return self._test_regression_score(mean_absolute_error, actual_slice, model, threshold)

    def test_r2(self, actual_slice, model: GiskardModel, threshold=1.0):
        """
        Test if the model R-Squared is higher than a threshold

        Example: The test is passed when the R-Squared is higher than 0.7

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for R-Squared

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            metric:
              The R-Squared metric
            passed:
              TRUE if R-Squared metric >= threshold
        """
        return self._test_regression_score(r2_score, actual_slice, model, threshold, r2=True)

    def _test_diff_prediction(
        self, test_fn, model, actual_slice, reference_slice, threshold=0.5, test_name=None
    ):
        self.do_save_results = False
        metric_1 = test_fn(reference_slice, model).metric
        metric_2 = test_fn(actual_slice, model).metric
        self.do_save_results = True
        try:
            change_pct = abs(metric_1 - metric_2) / metric_1
        except ZeroDivisionError:
            raise ZeroDivisionError(
                f"Unable to calculate performance difference: the {test_name} inside the"
                f" reference_slice is equal to zero"
            )

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                reference_slices_size=[len(reference_slice)],
                metric=change_pct,
                passed=change_pct < threshold,
            )
        )

    def test_diff_accuracy(self, actual_slice, reference_slice, model, threshold=0.1):
        """

        Test if the absolute percentage change of model Accuracy between two samples is lower than a threshold

        Example : The test is passed when the Accuracy for females has a difference lower than 10% from the
        Accuracy for males. For example, if the Accuracy for males is 0.8 (actual_slice) and the Accuracy  for
        females is 0.6 (reference_slice) then the absolute percentage Accuracy change is 0.2 / 0.8 = 0.25
        and the test will fail

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            reference_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Accuracy Score difference
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The Accuracy difference  metric
            passed:
              TRUE if Accuracy difference < threshold
        """
        return self._test_diff_prediction(
            self.test_accuracy,
            model,
            actual_slice,
            reference_slice,
            threshold,
            test_name="Accuracy",
        )

    def test_diff_f1(self, actual_slice, reference_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change in model F1 Score between two samples is lower than a threshold

        Example : The test is passed when the F1 Score for females has a difference lower than 10% from the
        F1 Score for males. For example, if the F1 Score for males is 0.8 (actual_slice) and the F1 Score  for
        females is 0.6 (reference_slice) then the absolute percentage F1 Score  change is 0.2 / 0.8 = 0.25
        and the test will fail

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            reference_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for F1 Score difference

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The F1 Score difference  metric
            passed:
              TRUE if F1 Score difference < threshold
        """
        return self._test_diff_prediction(
            self.test_f1, model, actual_slice, reference_slice, threshold, test_name="F1 Score"
        )

    def test_diff_precision(self, actual_slice, reference_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change of model Precision between two samples is lower than a threshold

        Example : The test is passed when the Precision for females has a difference lower than 10% from the
        Accuracy for males. For example, if the Precision for males is 0.8 (actual_slice) and the Precision  for
        females is 0.6 (reference_slice) then the absolute percentage Precision change is 0.2 / 0.8 = 0.25
        and the test will fail

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            reference_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Precision difference
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The Precision difference  metric
            passed:
              TRUE if Precision difference < threshold
        """
        return self._test_diff_prediction(
            self.test_precision,
            model,
            actual_slice,
            reference_slice,
            threshold,
            test_name="Precision",
        )

    def test_diff_recall(self, actual_slice, reference_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change of model Recall between two samples is lower than a threshold

        Example : The test is passed when the Recall for females has a difference lower than 10% from the
        Accuracy for males. For example, if the Recall for males is 0.8 (actual_slice) and the Recall  for
        females is 0.6 (reference_slice) then the absolute percentage Recall change is 0.2 / 0.8 = 0.25
        and the test will fail

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            reference_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Recall difference
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The Recall difference  metric
            passed:
              TRUE if Recall difference < threshold
        """
        return self._test_diff_prediction(
            self.test_recall, model, actual_slice, reference_slice, threshold, test_name="Recall"
        )

    def test_diff_reference_actual_f1(self, reference_slice, actual_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change in model F1 Score between reference and actual data
        is lower than a threshold

        Example : The test is passed when the F1 Score for reference dataset has a difference lower than 10% from the
        F1 Score for actual dataset. For example, if the F1 Score for reference dataset is 0.8 (reference_slice) and the
         F1 Score  for actual dataset is 0.6 (actual_slice) then the absolute percentage F1 Score  change is
        0.2 / 0.8 = 0.25 and the test will fail.

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            reference_slice(GiskardDataset):
              Slice of reference dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for F1 Score difference
        Returns:
          actual_slices_size:
              Length of actual_slice tested
          reference_slices_size:
              Length of reference_slice tested
          metric:
              The F1 Score difference  metric
          passed:
              TRUE if F1 Score difference < threshold
        """
        return self._test_diff_prediction(
            self.test_f1, model, reference_slice, actual_slice, threshold, test_name="F1 Score"
        )

    def test_diff_reference_actual_accuracy(
        self, reference_slice, actual_slice, model, threshold=0.1
    ):
        """
        Test if the absolute percentage change in model Accuracy between reference and actual data
        is lower than a threshold

        Example : The test is passed when the Accuracy for reference dataset has a difference lower than 10% from the
        Accuracy for actual dataset. For example, if the Accuracy for reference dataset is 0.8 (reference_slice) and the
         Accuracy  for actual dataset is 0.6 (actual_slice) then the absolute percentage Accuracy
        change is 0.2 / 0.8 = 0.25 and the test will fail.

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            reference_slice(GiskardDataset):
              Slice of reference dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for Accuracy difference
        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The Accuracy difference  metric
            passed:
              TRUE if Accuracy difference < threshold
        """
        return self._test_diff_prediction(
            self.test_accuracy,
            model,
            reference_slice,
            actual_slice,
            threshold,
            test_name="Accuracy",
        )

    def test_diff_rmse(self, actual_slice, reference_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change of model RMSE between two samples is lower than a threshold

        Example : The test is passed when the RMSE for females has a difference lower than 10% from the
        RMSE for males. For example, if the RMSE for males is 0.8 (actual_slice) and the RMSE  for
        females is 0.6 (reference_slice) then the absolute percentage RMSE change is 0.2 / 0.8 = 0.25
        and the test will fail

        Args:
            actual_slice(GiskardDataset):
              Slice of the actual dataset
            reference_slice(GiskardDataset):
              Slice of the actual dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for RMSE difference

        Returns:
            actual_slices_size:
              Length of actual_slice tested
            reference_slices_size:
              Length of reference_slice tested
            metric:
              The RMSE difference  metric
            passed:
              TRUE if RMSE difference < threshold
        """
        return self._test_diff_prediction(
            self.test_rmse, model, actual_slice, reference_slice, threshold, test_name="RMSE"
        )

    def test_diff_reference_actual_rmse(self, reference_slice, actual_slice, model, threshold=0.1):
        """
        Test if the absolute percentage change in model RMSE between reference and actual data
        is lower than a threshold

        Example : The test is passed when the RMSE for reference dataset has a difference lower than 10% from the
        RMSE for actual dataset. For example, if the RMSE for reference dataset is 0.8 (reference_slice) and the RMSE
        for actual dataset is 0.6 (actual_slice) then the absolute percentage RMSE  change is 0.2 / 0.8 = 0.25
        and the test will fail.

        Args:
            actual_slice(GiskardDataset):
              Slice of actual dataset
            reference_slice(GiskardDataset):
              Slice of reference dataset
            model(GiskardModel):
              Model used to compute the test
            threshold(float):
              Threshold value for RMSE difference
        Returns:
          actual_slices_size:
              Length of actual_slice tested
          reference_slices_size:
              Length of reference_slice tested
          metric:
              The RMSE difference  metric
          passed:
              TRUE if RMSE difference < threshold
        """
        return self._test_diff_prediction(
            self.test_rmse, model, reference_slice, actual_slice, threshold, test_name="RMSE"
        )



from typing import List

import inspect
import logging
from abc import ABC

from giskard.ml_worker.generated.ml_worker_pb2 import NamedSingleTestResult, SingleTestResult

logger = logging.getLogger(__name__)


class AbstractTestCollection(ABC):
    tests_results: List[NamedSingleTestResult]
    do_save_results = True

    def __init__(self, test_results: List[NamedSingleTestResult]) -> None:
        self.tests_results = test_results

    @staticmethod
    def _find_caller_test_name():
        stack = inspect.stack()
        for frame in stack:
            if frame.function.startswith("test_"):
                return frame.function
        logger.warning(f"Failed to extract test method name")

    def save_results(self, result: SingleTestResult, test_name=None):
        if not self.do_save_results:
            return result
        if test_name is None:
            test_name = self._find_caller_test_name()

        self.tests_results.append(NamedSingleTestResult(name=test_name, result=result))
        return result



import hashlib
import os

from mixpanel import Consumer, Mixpanel


def nofail(func):
    def inner_function(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except BaseException:  # NOSONAR
            pass

    return inner_function


def anonymize(message):
    if not message:
        return None
    if isinstance(message, list):
        return [anonymize(m) for m in message]

    return hashlib.sha1(str(message).encode()).hexdigest()[:10]


class GiskardAnalyticsCollector:
    def __init__(self) -> None:
        self.mp = self.configure_mixpanel()
        self.is_enabled = False
        self.distinct_user_id = None

    @staticmethod
    @nofail
    def configure_mixpanel():
        is_dev_mode = os.environ.get("GISKARD_DEV_MODE", "n").lower() in ["yes", "true", "1"]
        dev_mp_project_key = "4cca5fabca54f6df41ea500e33076c99"
        prod_mp_project_key = "2c3efacc6c26ffb991a782b476b8c620"

        return Mixpanel(
            dev_mp_project_key if is_dev_mode else prod_mp_project_key,
            consumer=Consumer(api_host="pxl.giskard.ai"),
        )

    @nofail
    def init(self, server_settings):
        giskard_instance_id = server_settings.get("app").get("generalSettings").get("instanceId")
        self.is_enabled = (
            server_settings.get("app").get("generalSettings").get("isAnalyticsEnabled")
        )
        user_login = server_settings.get("user").get("user_id")
        anonymous_login = anonymize(user_login)
        self.distinct_user_id = f"{giskard_instance_id}-{anonymous_login}"

        if self.is_enabled:
            self.mp.people_set(
                self.distinct_user_id,
                {
                    "Giskard Instance": giskard_instance_id,
                    "Giskard Version": server_settings.get("app").get("version"),
                },
            )

    @nofail
    def track(self, event_name, properties=None, meta=None):
        if self.is_enabled:
            self.mp.track(
                distinct_id=self.distinct_user_id,
                event_name=event_name,
                properties=properties,
                meta=meta,
            )



import pandas as pd

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult
from giskard.ml_worker.testing.abstract_test_collection import AbstractTestCollection


class HeuristicTests(AbstractTestCollection):
    def test_right_label(
        self,
        actual_slice: GiskardDataset,
        model: GiskardModel,
        classification_label: str,
        threshold=0.5,
    ) -> SingleTestResult:
        """
        Summary: Test if the model returns the right classification label for a slice

        Description: The test is passed when the percentage of rows returning the right
        classification label is higher than the threshold in a given slice

        Example: For a credit scoring model, the test is passed when more than 50%
        of people with high-salaries are classified as “non default”

        Args:
           actual_slice(GiskardDataset):
              Slice of the  actual dataset
          model(GiskardModel):
              Model used to compute the test
          classification_label(str):
              Classification label you want to test
          threshold(float):
              Threshold for the percentage of passed rows

        Returns:
          actual_slices_size:
              Length of actual_slice tested
          metrics:
              The ratio of rows with the right classification label over the total of rows in the slice
          passed:
              TRUE if passed_ratio > threshold
        """
        actual_slice.df.reset_index(drop=True, inplace=True)
        prediction_results = model.run_predict(actual_slice).prediction
        assert (
            classification_label in model.classification_labels
        ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'

        passed_idx = actual_slice.df.loc[prediction_results == classification_label].index.values

        passed_ratio = len(passed_idx) / len(actual_slice)
        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio > threshold,
            )
        )

    def test_output_in_range(
        self,
        actual_slice: GiskardDataset,
        model: GiskardModel,
        classification_label=None,
        min_range: float = 0.3,
        max_range: float = 0.7,
        threshold=0.5,
    ) -> SingleTestResult:
        """
        Summary: Test if the model output belongs to the right range for a slice

        Description: - The test is passed when the ratio of rows in the right range inside the
        slice is higher than the threshold.

         For classification: Test if the predicted probability for a given classification label
         belongs to the right range for a dataset slice

        For regression : Test if the predicted output belongs to the right range for a dataset slice

        Example :
        For Classification: For a credit scoring model, the test is passed when more than 50% of
        people with high wage have a probability of defaulting between 0 and 0.1

        For Regression : The predicted Sale Price of a house in the city falls in a particular range
        Args:
            actual_slice(GiskardDataset):
                Slice of the actual dataset
            model(GiskardModel):
                Model used to compute the test
            classification_label(str):
                Optional. Classification label you want to test
            min_range(float):
                Minimum probability of occurrence of classification label
            max_range(float):
                Maximum probability of occurrence of classification label
            threshold(float):
                Threshold for the percentage of passed rows

        Returns:
            actual_slices_size:
                Length of actual_slice tested
            metrics:
                The proportion of rows in the right range inside the slice
            passed:
                TRUE if metric > threshold
        """
        results_df = pd.DataFrame()
        actual_slice.df.reset_index(drop=True, inplace=True)

        prediction_results = model.run_predict(actual_slice)

        if model.model_type == "regression":
            results_df["output"] = prediction_results.raw_prediction

        elif model.model_type == "classification":
            assert (
                classification_label in model.classification_labels
            ), f'"{classification_label}" is not part of model labels: {",".join(model.classification_labels)}'
            results_df["output"] = prediction_results.all_predictions[classification_label]

        else:
            raise ValueError(f"Prediction task is not supported: {model.model_type}")

        passed_idx = actual_slice.df.loc[
            (results_df["output"] <= max_range) & (results_df["output"] >= min_range)
        ].index.values

        passed_ratio = len(passed_idx) / len(actual_slice)

        return self.save_results(
            SingleTestResult(
                actual_slices_size=[len(actual_slice)],
                metric=passed_ratio,
                passed=passed_ratio >= threshold,
            )
        )



import json
import warnings
from typing import Callable, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
import requests
from pandas.api.types import is_string_dtype
from requests_toolbelt.sessions import BaseUrlSession

from giskard.client.analytics_collector import GiskardAnalyticsCollector, anonymize
from giskard.client.io_utils import compress, pickle_dumps, save_df
from giskard.client.model import SupportedColumnType, SupportedModelTypes
from giskard.client.python_utils import get_python_requirements, get_python_version


class GiskardProject:
    def __init__(
            self, session: BaseUrlSession, project_key: str, analytics: GiskardAnalyticsCollector = None
    ) -> None:
        self.project_key = project_key
        self._session = session
        self.url = self._session.base_url.replace("/api/v2/", "")
        self.analytics = analytics or GiskardAnalyticsCollector()

    @staticmethod
    def _serialize(
            prediction_function: Callable[
                [pd.DataFrame],
                Iterable[Union[str, float, int]],
            ]
    ) -> bytes:
        compressed_pickle: bytes = compress(pickle_dumps(prediction_function))
        return compressed_pickle

    def upload_model(
            self,
            prediction_function: Callable[[pd.DataFrame], Iterable[Union[str, float, int]]],
            model_type: str,
            feature_names: List[str] = None,
            name: str = None,
            validate_df: pd.DataFrame = None,
            target: Optional[List[str]] = None,
            classification_threshold: Optional[float] = None,
            classification_labels: Optional[List[str]] = None,
    ):
        """
        Function to upload the Model to Giskard
        Args:
            prediction_function:
                The model you want to predict. It could be any Python function with the signature of
                predict_proba for classification: It returns the classification probabilities for all
                the classification labels
                predict for regression : It returns the predicted values for regression models.
            model_type:
                "classification" for classification model
                "regression" for regression model
            feature_names:
                 A list of the feature names of prediction_function. If provided, this list will be used to filter
                 the dataframe's columns before applying the model. By default, the dataframe is used as-is, meaning
                 that all of its columns are passed to the model.
                 Some important remarks:
                    - Make sure these features are contained in df
                    - Make sure that prediction_function(df[feature_names]) does not return an error message
                    - Make sure these features have the same order as the ones used in the
                      pipeline of prediction_function.
            name:
                The name of the model you want to upload
            validate_df:
                Dataset used to validate the model
            target:
                The column name in validate_df corresponding to the actual target variable (ground truth).
            classification_threshold:
                The probability threshold in the case of a binary classification model
            classification_labels:
                The classification labels of your prediction when prediction_task="classification".
                 Some important remarks:
                    - If classification_labels is a list of n elements, make sure prediction_function is
                     also returning probabilities
                    - Make sure the labels have the same order as the output of prediction_function
                    - Prefer using categorical values instead of numeric values in classification_labels
        """
        classification_labels, model = self._validate_model(
            classification_labels,
            classification_threshold,
            feature_names,
            model_type,
            prediction_function,
            target,
            validate_df,
        )
        self._post_model(
            classification_labels, classification_threshold, feature_names, model, model_type, name
        )

    def _post_model(
            self,
            classification_labels,
            classification_threshold,
            feature_names,
            model,
            model_type,
            name,
    ):
        requirements = get_python_requirements()
        params = {
            "name": name,
            "projectKey": self.project_key,
            "languageVersion": get_python_version(),
            "modelType": model_type,
            "threshold": classification_threshold,
            "featureNames": feature_names,
            "language": "PYTHON",
            "classificationLabels": classification_labels,
        }
        files = [
            ("metadata", (None, json.dumps(params), "application/json")),
            ("modelFile", model),
            ("requirementsFile", requirements),
        ]
        self._session.post("project/models/upload", data={}, files=files)
        self.analytics.track(
            "Upload Model",
            {
                "name": anonymize(name),
                "projectKey": anonymize(self.project_key),
                "languageVersion": get_python_version(),
                "modelType": model_type,
                "threshold": classification_threshold,
                "featureNames": anonymize(feature_names),
                "language": "PYTHON",
                "classificationLabels": anonymize(classification_labels),
            },
        )
        print(
            f"Model successfully uploaded to project key '{self.project_key}' and is available at {self.url} "
        )

    def _validate_model(
            self,
            classification_labels,
            classification_threshold,
            feature_names,
            model_type,
            prediction_function,
            target,
            validate_df,
    ):
        transformed_pred_func = self.transform_prediction_function(
            prediction_function, feature_names
        )

        self._validate_prediction_function(prediction_function)
        self._validate_model_type(model_type)
        classification_labels = self._validate_classification_labels(
            classification_labels, model_type
        )

        if model_type == SupportedModelTypes.CLASSIFICATION.value:
            self._validate_classification_threshold_label(
                classification_labels, classification_threshold
            )

        assert feature_names is None or isinstance(
            feature_names, list
        ), "Invalid feature_names parameter. Please provide the feature names as a list."

        if validate_df is not None:
            self._verify_is_pandasdataframe(validate_df)
            self._validate_features(feature_names=feature_names, validate_df=validate_df)

            if model_type == SupportedModelTypes.REGRESSION.value:
                self._validate_model_execution(
                    transformed_pred_func, validate_df, model_type, target=target
                )
            elif target is not None and model_type == SupportedModelTypes.CLASSIFICATION.value:
                self._validate_target(target, validate_df.keys())
                target_values = validate_df[target].unique()
                self._validate_label_with_target(classification_labels, target_values)
                self._validate_model_execution(
                    transformed_pred_func,
                    validate_df,
                    model_type,
                    classification_labels,
                    target=target,
                )
            else:  # Classification with target = None
                self._validate_model_execution(
                    transformed_pred_func,
                    validate_df,
                    model_type,
                    classification_labels,
                    target=target,
                )

        model = self._serialize(transformed_pred_func)
        return classification_labels, model

    def upload_df(
            self,
            df: pd.DataFrame,
            column_types: Dict[str, str],
            target: str = None,
            name: str = None,
    ) -> requests.Response:
        """
        Function to upload Dataset to Giskard
        Args:
            df:
                Dataset you want to upload
            column_types:
                A dictionary of column names and their types (numeric, category or text) for all columns of df.
            target:
                The column name in df corresponding to the actual target variable (ground truth).
            name:
                The name of the dataset you want to upload
        Returns:
                Response of the upload
        """
        data, raw_column_types = self._validate_and_compress_data(column_types, df, target)
        result = self._post_data(column_types, data, name, raw_column_types, target)
        return result

    def _post_data(self, column_types, data, name, raw_column_types, target):
        params = {
            "projectKey": self.project_key,
            "name": name,
            "featureTypes": column_types,
            "columnTypes": raw_column_types,
            "target": target,
        }
        files = [("metadata", (None, json.dumps(params), "application/json")), ("file", data)]
        result = self._session.post("project/data/upload", data={}, files=files)
        print(
            f"Dataset successfully uploaded to project key '{self.project_key}' and is available at {self.url} "
        )
        self.analytics.track(
            "Upload dataset",
            {
                "projectKey": anonymize(self.project_key),
                "name": anonymize(name),
                "featureTypes": anonymize(column_types),
                "target": anonymize(target),
            },
        )
        return result

    def _validate_and_compress_data(self, column_types, df, target):
        self._verify_is_pandasdataframe(df)
        if target is not None:
            self._validate_target(target, df.keys())
        self.validate_columns_columntypes(df, column_types)
        self._validate_column_types(column_types)
        self._verify_category_columns(df, column_types)
        raw_column_types = df.dtypes.apply(lambda x: x.name).to_dict()
        data = compress(save_df(df))
        return data, raw_column_types

    def upload_model_and_df(
            self,
            prediction_function: Callable[[pd.DataFrame], Iterable[Union[str, float, int]]],
            model_type: str,
            df: pd.DataFrame,
            column_types: Dict[str, str],
            feature_names: List[str] = None,
            target: str = None,
            model_name: str = None,
            dataset_name: str = None,
            classification_threshold: Optional[float] = None,
            classification_labels: Optional[List[str]] = None,
    ) -> None:
        """
        Function to upload Dataset and model to Giskard
        Args:
            prediction_function:
                The model you want to predict. It could be any Python function with the signature of
                predict_proba for classification: It returns the classification probabilities for all
                the classification labels
                predict for regression : It returns the predicted values for regression models.
            model_type:
                "classification" for classification model
                "regression" for regression model
            df:
                Dataset you want to upload
            column_types:
                A dictionary of column names and their types (numeric, category or text) for all columns of df.
            feature_names:
                 A list of the feature names of prediction_function. If provided, this list will be used to filter
                 the dataframe's columns before applying the model. By default, the dataframe is used as-is, meaning
                 that all of its columns are passed to the model.
                 Some important remarks:
                    - Make sure these features are contained in df
                    - Make sure that prediction_function(df[feature_names]) does not return an error message
                    - Make sure these features have the same order as the ones used
                      in the pipeline of prediction_function.
            target:
                The column name in df corresponding to the actual target variable (ground truth).
            model_name:
                The name of the model you want to upload
            dataset_name:
                The name of the dataset you want to upload
            classification_threshold:
                The probability threshold in the case of a binary classification model
            classification_labels:
                The classification labels of your prediction when prediction_task="classification".
                 Some important remarks:
                    - If classification_labels is a list of n elements, make sure prediction_function is
                     also returning probabilities
                    - Make sure the labels have the same order as the output of prediction_function
                    - Prefer using categorical values instead of numeric values in classification_labels
        """
        self.analytics.track("Upload model and dataset")
        data, raw_column_types = self._validate_and_compress_data(column_types, df, target)
        classification_labels, model = self._validate_model(
            classification_labels,
            classification_threshold,
            feature_names,
            model_type,
            prediction_function,
            target,
            df,
        )
        self._post_data(column_types, data, dataset_name, raw_column_types, target)
        self._post_model(
            classification_labels,
            classification_threshold,
            feature_names,
            model,
            model_type,
            model_name,
        )

    @staticmethod
    def _validate_model_type(model_type):
        if model_type not in {task.value for task in SupportedModelTypes}:
            raise ValueError(
                f"Invalid model_type parameter: {model_type}. "
                + f"Please choose one of {[task.value for task in SupportedModelTypes]}."
            )

    @staticmethod
    def _validate_column_types(column_types):
        if column_types and isinstance(column_types, dict):
            if not set(column_types.values()).issubset(
                    set(column_type.value for column_type in SupportedColumnType)
            ):
                raise ValueError(
                    f"Invalid column_types parameter: "
                    + f"Please choose types among {[column_type.value for column_type in SupportedColumnType]}."
                )
        else:
            raise ValueError(
                f"Invalid column_types parameter: {column_types}. Please specify non-empty dictionary."
            )

    @staticmethod
    def transform_prediction_function(prediction_function, feature_names=None):
        if feature_names:
            return lambda df: prediction_function(df[feature_names])
        else:
            return prediction_function

    @staticmethod
    def _validate_prediction_function(prediction_function):
        if not callable(prediction_function):
            raise ValueError(
                f"Invalid prediction_function parameter: {prediction_function}. Please specify Python function."
            )

    @staticmethod
    def _validate_target(target, dataframe_keys):
        if target not in dataframe_keys:
            raise ValueError(
                f"Invalid target parameter:"
                f" {target} column is not present in the dataset with columns:  {dataframe_keys}"
            )

    @staticmethod
    def _validate_features(feature_names=None, validate_df=None):
        if (
                feature_names is not None
                and validate_df is not None
                and not set(feature_names).issubset(set(validate_df.columns))
        ):
            missing_feature_names = set(feature_names) - set(validate_df.columns)
            raise ValueError(
                f"Value mentioned in  feature_names is  not available in validate_df: {missing_feature_names} "
            )

    @staticmethod
    def _validate_classification_threshold_label(
            classification_labels, classification_threshold=None
    ):
        if classification_labels is None:
            raise ValueError(f"Missing classification_labels parameter for classification model.")
        if classification_threshold is not None and not isinstance(
                classification_threshold, (int, float)
        ):
            raise ValueError(
                f"Invalid classification_threshold parameter: {classification_threshold}. Please specify valid number."
            )

        if classification_threshold is not None:
            if classification_threshold != 0.5:
                if len(classification_labels) != 2:
                    raise ValueError(
                        f"Invalid classification_threshold parameter: {classification_threshold} value is applicable "
                        f"only for binary classification. "
                    )

    @staticmethod
    def _validate_label_with_target(classification_labels, target_values=None):
        if target_values is not None:
            if not is_string_dtype(target_values):
                print(
                    'Hint: "Your target variable values are numeric. '
                    "It is recommended to have Human readable string as your target values "
                    'to make results more understandable in Giskard."'
                )

            target_values = (
                target_values
                if is_string_dtype(target_values)
                else [str(label) for label in target_values]
            )
            if not set(target_values).issubset(set(classification_labels)):
                invalid_target_values = set(target_values) - set(classification_labels)
                raise ValueError(
                    f"Target column value {invalid_target_values} not declared in "
                    f"classification_labels list: {classification_labels}"
                )

    @staticmethod
    def _validate_classification_labels(classification_labels, model_type):
        res = None
        if model_type == SupportedModelTypes.CLASSIFICATION.value:
            if classification_labels is not None and isinstance(
                    classification_labels, Iterable
            ):  # type: ignore
                if len(classification_labels) > 1:
                    res: Optional[List[str]] = [str(label) for label in classification_labels]
                else:
                    raise ValueError(
                        f"Invalid classification_labels parameter: {classification_labels}. "
                        f"Please specify more than 1 label."
                    )
            else:
                raise ValueError(
                    f"Invalid classification_labels parameter: {classification_labels}. "
                    f"Please specify valid list of strings."
                )
        if model_type == SupportedModelTypes.REGRESSION.value and classification_labels is not None:
            warnings.warn("'classification_labels' parameter is ignored for regression model")
            res = None
        return res

    @staticmethod
    def _validate_model_execution(
            prediction_function, df: pd.DataFrame, model_type, classification_labels=None, target=None
    ) -> None:
        try:
            if target is not None and target in df.columns:
                df = df.drop(target, axis=1)
            prediction = prediction_function(df)
        except Exception:
            raise ValueError(
                "Invalid prediction_function input.\n"
                "Please make sure that prediction_function(df[feature_names]) does not return an error "
                "message before uploading in Giskard"
            )
        GiskardProject._verify_prediction_output(df, model_type, prediction)
        if model_type == SupportedModelTypes.CLASSIFICATION.value:
            GiskardProject._validate_classification_prediction(classification_labels, prediction)

    @staticmethod
    def _verify_prediction_output(df: pd.DataFrame, model_type, prediction):
        assert len(df) == len(prediction), (
            f"Number of rows ({len(df)}) of dataset provided does not match with the "
            f"number of rows ({len(prediction)}) of prediction_function output"
        )
        if isinstance(prediction, np.ndarray) or isinstance(prediction, list):
            if model_type == SupportedModelTypes.CLASSIFICATION.value:
                if not any(isinstance(y, (np.floating, float)) for x in prediction for y in x):
                    raise ValueError("Model prediction should return float values ")
            if model_type == SupportedModelTypes.REGRESSION.value:
                if not any(isinstance(x, (np.floating, float)) for x in prediction):
                    raise ValueError("Model prediction should return float values ")
        else:
            raise ValueError("Model should return numpy array or a list")

    @staticmethod
    def _validate_classification_prediction(classification_labels, prediction):
        if not np.all(np.logical_and(prediction >= 0, prediction <= 1)):
            raise ValueError(
                "Invalid Classification Model prediction. Output probabilities should be in range [0,1]"
            )
        if not np.all(np.isclose(np.sum(prediction, axis=1), 1, atol=0.0000001)):
            raise ValueError(
                "Invalid Classification Model prediction. Sum of all probabilities should be 1 "
            )
        if prediction.shape[1] != len(classification_labels):
            raise ValueError(
                "Prediction output label shape and classification_labels shape do not match"
            )

    @staticmethod
    def validate_columns_columntypes(df: pd.DataFrame, column_types) -> pd.DataFrame:
        if not set(column_types.keys()).issubset(set(df.columns)):
            missing_columns = set(column_types.keys()) - set(df.columns)
            raise ValueError(
                f"Missing columns in dataframe according to column_types: {missing_columns}"
            )
        elif not set(df.columns).issubset(set(column_types.keys())):
            missing_columns = set(df.columns) - set(column_types.keys())
            raise ValueError(
                f"Invalid column_types parameter: Please declare the type for "
                f"{missing_columns} columns"
            )
        else:
            pandas_inferred_column_types = df.dtypes.to_dict()
            for column, dtype in pandas_inferred_column_types.items():
                if (
                        column_types.get(column) == SupportedColumnType.NUMERIC.value
                        and dtype == "object"
                ):
                    try:
                        df[column] = df[column].astype(float)
                    except Exception as e:
                        raise ValueError(f"Failed to convert column '{column}' to float") from e
            return df

    @staticmethod
    def _verify_category_columns(df: pd.DataFrame, column_types):
        for name, types in column_types.items():
            if types == SupportedColumnType.CATEGORY.value and len(df[name].unique()) > 30:
                warnings.warn(
                    f"Categorical feature '{name}' contains {len(df[name].unique())} distinct values. If "
                    f"necessary use 'numeric' or 'text' in column_types instead"
                )

    @staticmethod
    def _verify_is_pandasdataframe(df):
        assert isinstance(df, pd.DataFrame), "Dataset provided is not a pandas dataframe"

    def __repr__(self) -> str:
        return f"GiskardProject(project_key='{self.project_key}')"



import pandas as pd

from giskard.ml_worker.generated.ml_worker_pb2 import SingleTestResult
from giskard.ml_worker.utils.logging import timer


def ge_result_to_test_result(result, passed=True) -> SingleTestResult:
    """
        Converts a result of Great Expectations to TestResultMessage - java/python bridge test result class
    :param passed: boolean flag containing result of a test
    :param result: Great Expectations result
    :return: TestResultMessage - a protobuf generated test result class
    """

    return SingleTestResult(
        passed=passed,
        actual_slices_size=[result["element_count"]],
        missing_count=result["missing_count"],
        missing_percent=result["missing_percent"],
        unexpected_count=result["unexpected_count"],
        unexpected_percent=result["unexpected_percent"],
        unexpected_percent_total=result["unexpected_percent_total"],
        unexpected_percent_nonmissing=result["unexpected_percent_nonmissing"],
        partial_unexpected_index_list=result["partial_unexpected_index_list"],
        unexpected_index_list=result["unexpected_index_list"],
    )


@timer("Perturb data")
def apply_perturbation_inplace(df: pd.DataFrame, perturbation_dict):
    modified_rows = []
    i = 0
    for idx, r in df.iterrows():
        added = False
        for pert_col, pert_func in perturbation_dict.items():
            original_value = r[pert_col]
            new_value = pert_func(r)
            if original_value != new_value and not added:
                added = True
                modified_rows.append(i)
                df.loc[idx, pert_col] = new_value
        i += 1
    return modified_rows



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['giskard',
 'giskard.client',
 'giskard.ml_worker',
 'giskard.ml_worker.bridge',
 'giskard.ml_worker.core',
 'giskard.ml_worker.exceptions',
 'giskard.ml_worker.generated',
 'giskard.ml_worker.server',
 'giskard.ml_worker.testing',
 'giskard.ml_worker.utils']

package_data = \
{'': ['*']}

install_requires = \
['alibi[shap]>=0.7.0,<0.8.0',
 'beautifulsoup4>=4.11.1,<5.0.0',
 'click>=8.1.3,<9.0.0',
 'cloudpickle>=2.1.0,<3.0.0',
 'eli5>=0.13.0,<0.14.0',
 'great-expectations>=0.15.17,<0.16.0',
 'grpcio-status>=1.46.3,<2.0.0',
 'grpcio>=1.46.3,<2.0.0',
 'importlib_metadata>=4.11.4,<5.0.0',
 'lockfile>=0.12.2,<0.13.0',
 'mixpanel>=4.9.0,<5.0.0',
 'numpy>=1.21.6,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'protobuf>=3.9.2,<4.0.0',
 'psutil>=5.9.2,<6.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'python-daemon>=2.3.1,<3.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.28.1,<3.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'setuptools>=65.4.1,<66.0.0',
 'tenacity>=8.1.0,<9.0.0',
 'tqdm>=4.64.1,<5.0.0',
 'zstandard>=0.15.2,<0.16.0']

entry_points = \
{'console_scripts': ['giskard = giskard.cli:cli']}

setup_kwargs = {
    'name': 'giskard',
    'version': '1.7.0a2',
    'description': 'Inspect your AI models visually, find bugs, give feedback 🕵️\u200d♀️ 💬',
    'long_description': '# Giskard Client\n\n<div align="center">\n\n[![Python Version](https://img.shields.io/pypi/pyversions/giskard-client.svg)](https://pypi.org/project/giskard-client/)\n[![Dependencies Status](https://img.shields.io/badge/dependencies-up%20to%20date-brightgreen.svg)](https://github.com/Giskard-AI/giskard-client/pulls?utf8=%E2%9C%93&q=is%3Apr%20author%3Aapp%2Fdependabot)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Security: bandit](https://img.shields.io/badge/security-bandit-green.svg)](https://github.com/PyCQA/bandit)\n[![Pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/Giskard-AI/giskard-client/blob/main/.pre-commit-config.yaml)\n[![Semantic Versions](https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--versions-e10079.svg)](https://github.com/Giskard-AI/giskard-client/releases)\n[![License](https://img.shields.io/github/license/Giskard-AI/giskard-client)](https://github.com/Giskard-AI/giskard-client/blob/main/LICENSE)\n\nInspect your AI models visually, find bugs, give feedback 🕵️\u200d♀️ 💬\n\n</div>\n\n## Very first steps\n\n### Initial\n1. Clone the project in a local directory\n\n2. Using `pyenv` setup a local python 3.7\n\n3. If you don\'t have `Poetry` installed run:\n\n```bash\nmake download-poetry\n```\n\n4. Initialize poetry and install `pre-commit` hooks:\n\n```bash\nmake install\n```\n\n5. Tests folder contain scripts that can upload data to your Giskard Backend. Run them using\n```bash \nmake test\n```\nOr individually:\n\n```bash \npoetry run pytest tests/model_inspector/test_upload_text_classification.py\n```\n\nMake sure you have setup the correct URL, as well as your Giskard API token in a `.env` file at the root of the project. \n\n### Poetry\n\nWant to know more about Poetry? Check [its documentation](https://python-poetry.org/docs/).\n\n<details>\n<summary>Details about Poetry</summary>\n<p>\n\nPoetry\'s [commands](https://python-poetry.org/docs/cli/#commands) are very intuitive and easy to learn, like:\n\n- `poetry add numpy@latest`\n- `poetry run pytest`\n- `poetry publish --build`\n\netc\n</p>\n</details>\n\n### Building and releasing your package\n\nBuilding a new version of the application contains steps:\n\n- Bump the version of your package `poetry version <version>`. You can pass the new version explicitly, or a rule such as `major`, `minor`, or `patch`. For more details, refer to the [Semantic Versions](https://semver.org/) standard.\n- Make a commit to `GitHub`.\n- Create a `GitHub release`.\n- And... publish 🙂 `poetry publish --build`\n\n## 🎯 What\'s next\n\nWell, that\'s up to you 💪🏻. I can only recommend the packages and articles that helped me.\n\n- [`Typer`](https://github.com/tiangolo/typer) is great for creating CLI applications.\n- [`Rich`](https://github.com/willmcgugan/rich) makes it easy to add beautiful formatting in the terminal.\n- [`Pydantic`](https://github.com/samuelcolvin/pydantic/) – data validation and settings management using Python type hinting.\n- [`Loguru`](https://github.com/Delgan/loguru) makes logging (stupidly) simple.\n- [`tqdm`](https://github.com/tqdm/tqdm) – fast, extensible progress bar for Python and CLI.\n- [`IceCream`](https://github.com/gruns/icecream) is a little library for sweet and creamy debugging.\n- [`orjson`](https://github.com/ijl/orjson) – ultra fast JSON parsing library.\n- [`Returns`](https://github.com/dry-python/returns) makes you function\'s output meaningful, typed, and safe!\n- [`Hydra`](https://github.com/facebookresearch/hydra) is a framework for elegantly configuring complex applications.\n- [`FastAPI`](https://github.com/tiangolo/fastapi) is a type-driven asynchronous web framework.\n\nArticles:\n\n- [Open Source Guides](https://opensource.guide/).\n- [A handy guide to financial support for open source](https://github.com/nayafia/lemonade-stand)\n- [GitHub Actions Documentation](https://help.github.com/en/actions).\n- Maybe you would like to add [gitmoji](https://gitmoji.carloscuesta.me/) to commit names. This is really funny. 😄\n\n## 🚀 Features\n\n### Development features\n\n- Supports for `Python 3.7` and higher.\n- [`Poetry`](https://python-poetry.org/) as the dependencies manager. See configuration in [`pyproject.toml`](https://github.com/Giskard-AI/giskard-client/blob/main/pyproject.toml) and [`setup.cfg`](https://github.com/Giskard-AI/giskard-client/blob/main/setup.cfg).\n- Automatic codestyle with [`black`](https://github.com/psf/black), [`isort`](https://github.com/timothycrosley/isort) and [`pyupgrade`](https://github.com/asottile/pyupgrade).\n- Ready-to-use [`pre-commit`](https://pre-commit.com/) hooks with code-formatting.\n- Type checks with [`mypy`](https://mypy.readthedocs.io); docstring checks with [`darglint`](https://github.com/terrencepreilly/darglint); security checks with [`safety`](https://github.com/pyupio/safety) and [`bandit`](https://github.com/PyCQA/bandit)\n- Testing with [`pytest`](https://docs.pytest.org/en/latest/).\n- Ready-to-use [`.editorconfig`](https://github.com/Giskard-AI/giskard-client/blob/main/.editorconfig), [`.dockerignore`](https://github.com/Giskard-AI/giskard-client/blob/main/.dockerignore), and [`.gitignore`](https://github.com/Giskard-AI/giskard-client/blob/main/.gitignore). You don\'t have to worry about those things.\n\n### Deployment features\n\n- `GitHub` integration: issue and pr templates.\n- `Github Actions` with predefined [build workflow](https://github.com/Giskard-AI/giskard-client/blob/main/.github/workflows/build.yml) as the default CI/CD.\n- Everything is already set up for security checks, codestyle checks, code formatting, testing, linting, docker builds, etc with [`Makefile`](https://github.com/Giskard-AI/giskard-client/blob/main/Makefile#L89). More details in [makefile-usage](#makefile-usage).\n- [Dockerfile](https://github.com/Giskard-AI/giskard-client/blob/main/docker/Dockerfile) for your package.\n- Always up-to-date dependencies with [`@dependabot`](https://dependabot.com/). You will only [enable it](https://docs.github.com/en/github/administering-a-repository/enabling-and-disabling-version-updates#enabling-github-dependabot-version-updates).\n- Automatic drafts of new releases with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter). You may see the list of labels in [`release-drafter.yml`](https://github.com/Giskard-AI/giskard-client/blob/main/.github/release-drafter.yml). Works perfectly with [Semantic Versions](https://semver.org/) specification.\n\n### Open source community features\n\n- Ready-to-use [Pull Requests templates](https://github.com/Giskard-AI/giskard-client/blob/main/.github/PULL_REQUEST_TEMPLATE.md) and several [Issue templates](https://github.com/Giskard-AI/giskard-client/tree/main/.github/ISSUE_TEMPLATE).\n- Files such as: `LICENSE`, `CONTRIBUTING.md`, `CODE_OF_CONDUCT.md`, and `SECURITY.md` are generated automatically.\n- [`Stale bot`](https://github.com/apps/stale) that closes abandoned issues after a period of inactivity. (You will only [need to setup free plan](https://github.com/marketplace/stale)). Configuration is [here](https://github.com/Giskard-AI/giskard-client/blob/main/.github/.stale.yml).\n- [Semantic Versions](https://semver.org/) specification with [`Release Drafter`](https://github.com/marketplace/actions/release-drafter).\n\n## Installation\n\n```bash\npip install -U giskard-client\n```\n\nor install with `Poetry`\n\n```bash\npoetry add giskard-client\n```\n\n\n\n### Makefile usage\n\n[`Makefile`](https://github.com/Giskard-AI/giskard-client/blob/main/Makefile) contains a lot of functions for faster development.\n\n<details>\n<summary>1. Download and remove Poetry</summary>\n<p>\n\nTo download and install Poetry run:\n\n```bash\nmake poetry-download\n```\n\nTo uninstall\n\n```bash\nmake poetry-remove\n```\n\n</p>\n</details>\n\n<details>\n<summary>2. Install all dependencies and pre-commit hooks</summary>\n<p>\n\nInstall requirements:\n\n```bash\nmake install\n```\n\nPre-commit hooks coulb be installed after `git init` via\n\n```bash\nmake pre-commit-install\n```\n\n</p>\n</details>\n\n<details>\n<summary>3. Codestyle</summary>\n<p>\n\nAutomatic formatting uses `pyupgrade`, `isort` and `black`.\n\n```bash\nmake codestyle\n\n# or use synonym\nmake formatting\n```\n\nCodestyle checks only, without rewriting files:\n\n```bash\nmake check-codestyle\n```\n\n> Note: `check-codestyle` uses `isort`, `black` and `darglint` library\n\n<details>\n<summary>4. Code security</summary>\n<p>\n\n```bash\nmake check-safety\n```\n\nThis command launches `Poetry` integrity checks as well as identifies security issues with `Safety` and `Bandit`.\n\n```bash\nmake check-safety\n```\n\n</p>\n</details>\n\n</p>\n</details>\n\n<details>\n<summary>5. Type checks</summary>\n<p>\n\nRun `mypy` static type checker\n\n```bash\nmake mypy\n```\n\n</p>\n</details>\n\n<details>\n<summary>6. Tests</summary>\n<p>\n\nRun `pytest`\n\n```bash\nmake test\n```\n\n</p>\n</details>\n\n<details>\n<summary>7. All linters</summary>\n<p>\n\nOf course there is a command to ~~rule~~ run all linters in one:\n\n```bash\nmake lint\n```\n\nthe same as:\n\n```bash\nmake test && make check-codestyle && make mypy && make check-safety\n```\n\n</p>\n</details>\n\n<details>\n<summary>8. Docker</summary>\n<p>\n\n```bash\nmake docker-build\n```\n\nwhich is equivalent to:\n\n```bash\nmake docker-build VERSION=latest\n```\n\nRemove docker image with\n\n```bash\nmake docker-remove\n```\n\nMore information [about docker](https://github.com/Giskard-AI/giskard-client/tree/main/docker).\n\n</p>\n</details>\n\n<details>\n<summary>9. Cleanup</summary>\n<p>\nDelete pycache files\n\n```bash\nmake pycache-remove\n```\n\nRemove package build\n\n```bash\nmake build-remove\n```\n\nOr to remove pycache, build and docker image run:\n\n```bash\nmake clean-all\n```\n\n</p>\n</details>\n\n## 📈 Releases\n\nYou can see the list of available releases on the [GitHub Releases](https://github.com/Giskard-AI/giskard-client/releases) page.\n\nWe follow [Semantic Versions](https://semver.org/) specification.\n\nWe use [`Release Drafter`](https://github.com/marketplace/actions/release-drafter). As pull requests are merged, a draft release is kept up-to-date listing the changes, ready to publish when you’re ready. With the categories option, you can categorize pull requests in release notes using labels.\n\n### List of labels and corresponding titles\n\n|               **Label**               |  **Title in Releases**  |\n| :-----------------------------------: | :---------------------: |\n|       `enhancement`, `feature`        |       🚀 Features       |\n| `bug`, `refactoring`, `bugfix`, `fix` | 🔧 Fixes & Refactoring  |\n|       `build`, `ci`, `testing`        | 📦 Build System & CI/CD |\n|              `breaking`               |   💥 Breaking Changes   |\n|            `documentation`            |    📝 Documentation     |\n|            `dependencies`             | ⬆️ Dependencies updates |\n\nYou can update it in [`release-drafter.yml`](https://github.com/Giskard-AI/giskard-client/blob/main/.github/release-drafter.yml).\n\nGitHub creates the `bug`, `enhancement`, and `documentation` labels for you. Dependabot creates the `dependencies` label. Create the remaining labels on the Issues tab of your GitHub repository, when you need them.\n\n## 🛡 License\n\n[![License](https://img.shields.io/github/license/Giskard-AI/giskard-client)](https://github.com/Giskard-AI/giskard-client/blob/main/LICENSE)\n\nThis project is licensed under the terms of the `Apache Software License 2.0` license. See [LICENSE](https://github.com/Giskard-AI/giskard-client/blob/main/LICENSE) for more details.\n\n## 📃 Citation\n\n```bibtex\n@misc{giskard-client,\n  author = {Giskard AI},\n  title = {Inspect your AI models visually, find bugs, give feedback 🕵️\u200d♀️ 💬},\n  year = {2021},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/Giskard-AI/giskard-client\n}\n```\n\n## Credits [![🚀 Your next Python package needs a bleeding-edge project structure.](https://img.shields.io/badge/python--package--template-%F0%9F%9A%80-brightgreen)](https://github.com/TezRomacH/python-package-template)\n\nThis project was generated with [`python-package-template`](https://github.com/TezRomacH/python-package-template)\n',
    'author': 'Giskard AI',
    'author_email': 'hello@giskard.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Giskard-AI/giskard-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



from typing import Callable, Iterable, Union

import functools
import logging
import traceback

import grpc
from google.protobuf import any_pb2
from google.protobuf.message import Message
from google.rpc.error_details_pb2 import DebugInfo
from google.rpc.status_pb2 import Status
from grpc import ServicerContext, StatusCode
from grpc.experimental import wrap_server_method_handler
from grpc_status import rpc_status

from giskard.ml_worker.exceptions.IllegalArgumentError import CodedError

MESSAGE_TYPE = Union[Message, Iterable[Message]]

logger = logging.getLogger(__name__)


class ErrorInterceptor(grpc.aio.ServerInterceptor):
    @staticmethod
    def terminate_with_exception(error_code: StatusCode, e: Exception, context: ServicerContext):
        detail = any_pb2.Any()
        detail.Pack(
            DebugInfo(
                stack_entries=traceback.format_stack(),
                detail=str(e),
            )
        )
        code, _ = error_code.value
        rich_status = Status(code=code, message=e.__class__.__name__, details=[detail])
        context.abort_with_status(rpc_status.to_status(rich_status))

    @staticmethod
    def _wrapper(
        behavior: Callable[[MESSAGE_TYPE, ServicerContext], MESSAGE_TYPE]
    ) -> Callable[[MESSAGE_TYPE, ServicerContext], Message]:
        @functools.wraps(behavior)
        def wrapper(request: MESSAGE_TYPE, context: ServicerContext) -> MESSAGE_TYPE:
            try:
                return behavior(request, context)
            except CodedError as e:
                logger.exception(e)
                ErrorInterceptor.terminate_with_exception(e.code, e, context)
            except Exception as e:
                logger.exception(e)
                ErrorInterceptor.terminate_with_exception(StatusCode.INTERNAL, e, context)

        return wrapper

    async def intercept_service(self, continuation, handler_call_details):
        return wrap_server_method_handler(
            ErrorInterceptor._wrapper, continuation(handler_call_details)
        )



"""API Client to interact with the Giskard app"""
from typing import List

import logging
import warnings
from urllib.parse import urljoin

from requests.adapters import HTTPAdapter
from requests_toolbelt import sessions

import giskard
from giskard.client.analytics_collector import GiskardAnalyticsCollector, anonymize
from giskard.client.project import GiskardProject


class GiskardError(Exception):
    def __init__(self, message: str, status: int, code: str) -> None:
        super().__init__(message)
        self.status = status
        self.code = code


class ErrorHandlingAdapter(HTTPAdapter):
    def build_response(self, req, resp):
        response = super().build_response(req, resp)

        if not response.ok:
            giskard_error = None
            try:
                err_resp = response.json()
                giskard_error = GiskardError(
                    status=err_resp.get("status"),
                    code=err_resp.get("message"),
                    message=f"{err_resp.get('title', 'Unknown error')}: {err_resp.get('detail', 'no details')}",
                )
            except:  # NOSONAR
                response.raise_for_status()
            raise giskard_error
        return response


class GiskardClient:
    def __init__(self, url: str, token: str):
        base_url = urljoin(url, "/api/v2/")
        self._session = sessions.BaseUrlSession(base_url=base_url)
        self._session.mount(base_url, ErrorHandlingAdapter())
        self._session.headers.update({"Authorization": f"Bearer {token}"})
        self.analytics = GiskardAnalyticsCollector()
        try:
            server_settings = self._session.get("settings").json()
            self.analytics.init(server_settings)
        except Exception:
            logging.warning(f"Failed to fetch server settings", exc_info=True)
        self.analytics.track("Init GiskardClient", {"client version": giskard.__version__})

    @property
    def session(self):
        return self._session

    def list_projects(self) -> List[GiskardProject]:
        self.analytics.track("List Projects")
        response = self._session.get("projects").json()
        return [GiskardProject(self._session, p["key"], analytics=self.analytics) for p in response]

    def get_project(self, project_key: str):
        """
        Function to get the project that belongs to the mentioned project key
        Args:
            project_key:
                The unique value of  project provided during project creation
        Returns:
            GiskardProject:
                The giskard project that belongs to the project key
        """
        self.analytics.track("Get Project", {"project_key": anonymize(project_key)})
        response = self._session.get(f"project", params={"key": project_key}).json()
        return GiskardProject(self._session, response["key"], analytics=self.analytics)

    def create_project(self, project_key: str, name: str, description: str = None):
        """
        Function to create a project in Giskard
        Args:
            project_key:
                The unique value of the project which will be used to identify  and fetch teh project in future
            name:
                The name of the project
            description:
                Describe your project
        Returns:
            GiskardProject:
                The project created in giskard
        """
        self.analytics.track(
            "Create Project",
            {
                "project_key": anonymize(project_key),
                "description": anonymize(description),
                "name": anonymize(name),
            },
        )
        try:
            response = self._session.post(
                "project", json={"description": description, "key": project_key, "name": name}
            ).json()
        except GiskardError as e:
            if e.code == "error.http.409":
                warnings.warn(
                    "This project key already exists. "
                    "If you want to reuse existing project use get_project(“project_key”) instead"
                )
            raise e
        actual_project_key = response.get("key")
        if actual_project_key != project_key:
            print(f"Project created with a key : {actual_project_key}")
        return GiskardProject(self._session, actual_project_key, analytics=self.analytics)



from io import BytesIO

import cloudpickle
import pandas as pd
from zstandard import ZstdDecompressor, decompress

from giskard.ml_worker.core.giskard_dataset import GiskardDataset
from giskard.ml_worker.core.model import GiskardModel
from giskard.ml_worker.generated.ml_worker_pb2 import (
    SerializedGiskardDataset,
    SerializedGiskardModel,
)
from giskard.ml_worker.utils.logging import timer


@timer()
def deserialize_model(serialized_model: SerializedGiskardModel) -> GiskardModel:
    deserialized_model = GiskardModel(
        cloudpickle.load(
            ZstdDecompressor().stream_reader(serialized_model.serialized_prediction_function)
        ),
        model_type=serialized_model.model_type,
        classification_threshold=serialized_model.threshold.value
        if serialized_model.HasField("threshold")
        else None,
        feature_names=list(serialized_model.feature_names),
        classification_labels=list(serialized_model.classification_labels),
    )

    return deserialized_model


@timer()
def deserialize_dataset(serialized_dataset: SerializedGiskardDataset) -> GiskardDataset:
    deserialized_dataset = GiskardDataset(
        df=pd.read_csv(
            BytesIO(decompress(serialized_dataset.serialized_df)),
            keep_default_na=False,
            na_values=["_GSK_NA_"],
        ),
        target=serialized_dataset.target,
        feature_types=dict(serialized_dataset.feature_types),
        column_types=dict(serialized_dataset.column_types),
    )

    return deserialized_dataset
